/* 
 * $Id: bgpconf.c,v 1.1 1998/09/01 19:39:28 masaki Exp $
 */

#include <ctype.h>
#include <mrt.h>
#include <config_file.h>
#include <protoconf.h>


static void
get_config_aspath_filter (int num)
{
    char *cp = as_access_list_toa (num);
    if (cp) {
	config_add_output ("%s", cp);
	Delete (cp);
    }
}


static int
config_ip_as_filter (uii_connection_t * uii, int num, char *permit_or_deny,
		     char *re)
{
    int permit = 0;

    if (strcasecmp ("permit", permit_or_deny) == 0)
	permit = 1;

    if (num < 0 && num >= MAX_AS_ALIST) {
	config_notice (TR_ERROR, uii,
		   "invalid as-path access-list number (%d)\n", num);
	Delete (permit_or_deny);
	Delete (re);
	return (-1);
    }

    if (add_as_access_list (num, re, permit) > 0) {
	config_notice (TR_TRACE, uii,
		       "CONFIG as-path access-list %d %s %s\n",
		       num, (permit) ? "permit" : "deny", re);
	config_add_module (0, "as-path access-list", get_config_aspath_filter,
			   (void *) num);
    }
    else {
        Delete (permit_or_deny);
        Delete (re);
        return (-1);
    }
    Delete (permit_or_deny);
    Delete (re);
    return (1);
}


static void
print_aggregate_address (prefix_t * prefix, u_long option, int view)
{
    char buff[MAXLINE];

    if (prefix->family == AF_INET && view != 0)
	sprintf (buff, " view %d", view);
#ifdef HAVE_IPV6
    else if (prefix->family == AF_INET6 && view != 1)
	sprintf (buff, " view %d", view);
#endif /* HAVE_IPV6 */
    else
	strcpy (buff, "");
    config_add_output ("  aggregate-address %s%s%s%s\n",
		       prefix_toax (prefix),
		       BIT_TEST (option, BGP_AGGOPT_AS_SET) ? " as-set" : "",
	  BIT_TEST (option, BGP_AGGOPT_SUMMARY_ONLY) ? " summary-only" : "",
		       buff);
}


static void
get_config_router_bgp ()
{
    bgp_peer_t *peer;
    int i;
    prefix_t *prefix;

    config_add_output ("router bgp %d\n", BGP->my_as);

    /* I think this is thread-safe... we're just reading data */
    /* NO. someone may change the peer's list, and it may break here */

    pthread_mutex_lock (&BGP->mutex_lock);

    LL_Iterate (BGP->ll_networks, prefix) {
	config_add_output ("  network %s\n", prefix_toax (prefix));
    }

    LL_Iterate (BGP->ll_bgp_peers, peer) {
	char name[MAXLINE];

        pthread_mutex_lock (&peer->mutex_lock);

	strcpy (name, (peer->name)? peer->name: prefix_toa (peer->peer_addr));

	if (peer->name) {
	    config_add_output ("  neighbor %s peer %s src %s iface %s\n", 
		peer->name, prefix_toa (peer->peer_addr), 
		prefix_toa (peer->bind_addr), peer->bind_if->name);
	}

	if (peer->gateway->AS > 0)
	    config_add_output ("  neighbor %s remote-as %d\n", name,
		               peer->gateway->AS);
	if (peer->dlist_in >= 0)
	    config_add_output ("  neighbor %s distribute-list %d in\n",
			name, peer->dlist_in);
	if (peer->dlist_out >= 0)
	    config_add_output ("  neighbor %s distribute-list %d out\n",
		       name, peer->dlist_out);
	if (peer->flist_in >= 0)
	    config_add_output ("  neighbor %s filter-list %d in\n",
			name, peer->flist_in);
	if (peer->flist_out >= 0)
	    config_add_output ("  neighbor %s filter-list %d out\n",
		       name, peer->flist_out);
	if (peer->route_map_in >= 0)
	    config_add_output ("  neighbor %s route-map %d in\n",
			name, peer->route_map_in);
	if (peer->route_map_out >= 0)
	    config_add_output ("  neighbor %s route-map %d out\n",
		       name, peer->route_map_out);

	if (peer->maximum_prefix > 0)
	    config_add_output ("  neighbor %s maximum-prefix %d\n",
		       peer->maximum_prefix);

	if (BIT_TEST (peer->options, BGP_CONNECT_PASSIVE))
	    config_add_output ("  neighbor %s passive\n",
			       name);
	if (BIT_TEST (peer->options, BGP_TRANSPARENT_AS))
	    config_add_output ("  neighbor %s transparent-as\n", name);
	if (BIT_TEST (peer->options, BGP_TRANSPARENT_NEXTHOP))
	    config_add_output ("  neighbor %s transparent-nexthop\n", name);
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_01))
	    config_add_output ("  neighbor %s bgp4+ 1\n", name);
	if (BIT_TEST (peer->options, BGP_BGP4PLUS_AUTO))
	    config_add_output ("  neighbor %s bgp4+ auto\n", name);

	if (peer->HoldTime_Interval >= 0)
	    config_add_output ("  neighbor %s holtime %d\n",
		               name, peer->HoldTime_Interval);
	if (peer->KeepAlive_Interval >= 0)
	    config_add_output ("  neighbor %s keepalive %d\n",
		               name, peer->KeepAlive_Interval);
	if (peer->ConnectRetry_Interval >= 0)
	    config_add_output ("  neighbor %s connectretry %d\n",
		               name, peer->ConnectRetry_Interval);
	if (peer->Start_Interval >= 0)
	    config_add_output ("  neighbor %s starttime %d\n",
		               name, peer->Start_Interval);

        pthread_mutex_unlock (&peer->mutex_lock);
    }
    pthread_mutex_unlock (&BGP->mutex_lock);

    pthread_mutex_lock (&MRT->mutex_lock);
    for (i = 0; i <= PROTO_MAX; i++) {
        if (BIT_TEST (MRT->redist[i], (1 << PROTO_BGP)))
	    config_add_output ("  redistribute %s\n", proto2string (i));
    }
    pthread_mutex_unlock (&MRT->mutex_lock);

    for (i = 0; i < MAX_BGP_VIEWS; i++) {
	view_print_aggregate (i, print_aggregate_address);
    }
}


static int
config_router_bgp (uii_connection_t * uii, int as)
{
    if (running && !BIT_TEST (MRT->protocols, 1 << PROTO_BGP)) {
        config_notice (TR_ERROR, uii, "You can not do this while running\n");
	return (0);
    }

    if (as <= 0 || (BGP->my_as != 0 && BGP->my_as != as)) {
	config_notice (TR_TRACE, uii,
		       "BGP is already configured; AS is %d\n", BGP->my_as);
	return (-1);
    }

    CONFIG_MRTD->protocol = PROTO_BGP;
    uii->state = UII_CONFIG_ROUTER_BGP;
    config_notice (TR_TRACE, uii, "Router BGP %d\n", as);

    if (BIT_TEST (MRT->protocols, 1 << PROTO_BGP)) {
	return (0);
    }

    set_BGP (BGP_MY_AS, as, NULL);
    config_add_module (CF_DELIM, "router bgp", get_config_router_bgp, NULL);

    /* probably OK without locking */
    pthread_mutex_lock (&MRT->mutex_lock);
    BIT_SET (MRT->protocols, 1 << PROTO_BGP);
    pthread_mutex_unlock (&MRT->mutex_lock);

    return (1);
}


/* config_router_neighbor_remoteas
 * neighbor %p remote-as %d
 */
static int
config_router_neighbor_remoteas (uii_connection_t * uii,
				 prefix_t * prefix, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	/* I am only one except for automatic addition that adds a peer */
	peer = Add_BGP_Peer (NULL, prefix, NULL, NULL, as, BGP->trace);
	assert (peer);
	if (BGPSIM_TRANSPARENT) {
	    /* BGPSIM should behave as transparent */
            pthread_mutex_lock (&peer->mutex_lock);
	    BIT_SET (peer->options, BGP_TRANSPARENT_AS);
	    BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
            pthread_mutex_unlock (&peer->mutex_lock);
	}
    }
    else {
	config_notice (TR_ERROR, uii, "Peer is already defined\n");
    	Deref_Prefix (prefix);
    	return (-1);
    }
    if (BGP->my_as == as) {
        pthread_mutex_lock (&peer->mutex_lock);
        pthread_mutex_unlock (&peer->mutex_lock);
	trace (TR_PARSE, BGP->trace, "neighbor %s AS%d as iBGP\n",
	       prefix_toa (prefix), as);
    }
    else {
	trace (TR_PARSE, BGP->trace, "neighbor %s AS%d as eBGP\n",
	       prefix_toa (prefix), as);
    }
    Deref_Prefix (prefix);
    if (running)
	start_bgp_peer (peer);
    return (1);
}


/* no_config_router_neighbor_remoteas
 * no neighbor %p remote-as %d
 */
static int
no_config_router_neighbor_remoteas (uii_connection_t * uii,
				    prefix_t * prefix, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer does not exist\n");
	Deref_Prefix (prefix);
	return (-1);
    }
    /* XXX */

    Destroy_BGP_Peer (peer);
    config_notice (NORM, uii, "Peer deleted\n");
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_peer (uii_connection_t * uii, char *name,
			prefix_t * prefix, char *option)
{
    bgp_peer_t *peer;
    prefix_t *usrc = NULL;
    char ifname[MAXLINE];
    interface_t *interface = NULL;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) != NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is already defined\n",
			prefix_toa (prefix));
    	Delete (name);
    	Deref_Prefix (prefix);
    	Delete (option);
    	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByID (name)) != NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is already defined\n", name);
    	Delete (name);
    	Deref_Prefix (prefix);
    	Delete (option);
    	return (-1);
    }

    strcpy (ifname, "");
    if (option) {
        if (parse_line (option, "src %m iface %s", &usrc, ifname) <= 0 &&
            parse_line (option, "iface %s src %m", ifname, &usrc) <= 0) {
    	    Delete (name);
    	    Deref_Prefix (prefix);
    	    Delete (option);
    	    return (-1);
	}
	Delete (option);
    }

    /* XXX usrc should be checked against interface addresses */

    if (ifname[0] && (interface = find_interface_byname (ifname)) == NULL) {
	config_notice (TR_ERROR, uii, "Interface %s is not found\n", ifname);
    	Delete (name);
    	Deref_Prefix (prefix);
    	if (usrc) Deref_Prefix (usrc);
    	return (-1);
    }

    /* I am only one except for automatic addition that adds a peer */
    peer = Add_BGP_Peer (name, prefix, usrc, interface, 0, BGP->trace);
    assert (peer);

	if (BGPSIM_TRANSPARENT) {
	    /* BGPSIM should behave as transparent */
            pthread_mutex_lock (&peer->mutex_lock);
	    BIT_SET (peer->options, BGP_TRANSPARENT_AS);
	    BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
            pthread_mutex_unlock (&peer->mutex_lock);
	}

    Delete (name);
    Deref_Prefix (prefix);
    if (usrc) Deref_Prefix (usrc);
    return (1);
}



static int
config_router_neighbor_n_remoteas (uii_connection_t * uii, char *name, int as)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer %s is not defined\n", name);
    	Delete (name);
    	return (-1);
    }

    if (BGP->my_as == as) {
        pthread_mutex_lock (&peer->mutex_lock);
        pthread_mutex_unlock (&peer->mutex_lock);
	trace (TR_PARSE, BGP->trace,
	       "neighbor %s AS%d as iBGP\n", name, as);
    }
    else {
	trace (TR_PARSE, BGP->trace,
	       "neighbor %s AS%d as eBGP\n", name, as);
    }
    Delete (name);
    if (running)
	start_bgp_peer (peer);
    return (1);
}


static int
no_config_router_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "Peer %s does not exist\n", name);
	Delete (name);
	return (-1);
    }
    /* XXX */

    Destroy_BGP_Peer (peer);
    config_notice (TR_INFO, uii, "Peer %s deleted\n", name);
    Delete (name);
    return (1);
}


/* 
 * config_router_neighbor_view
 * just used by BGP to configure peer (at least for now)
 */
static int
config_router_neighbor_view (uii_connection_t * uii, prefix_t * prefix,
			     int view)
{
    bgp_peer_t *peer;
    char str[MAXLINE];

    if (view >= MAX_BGP_VIEWS || view < 0) {
	config_notice (TR_ERROR, uii,
		       "Invalid view %d for BGP neighbor %s [0-%d]\n",
		       view, prefix_toa (prefix), MAX_BGP_VIEWS - 1);
	Deref_Prefix (prefix);
	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    if (BGP->views[view]) {
	if (BGP->views[view]->family != prefix->family) {
	    config_notice (TR_ERROR, uii, "view %d already defined for %s\n",
			   view, family2string (BGP->views[view]->family));
	    Deref_Prefix (prefix);
	    return (-1);
	}

	trace (TR_PARSE, BGP->trace, "neighbor %s view %d\n",
	       prefix_toa (prefix), view);

        pthread_mutex_lock (&MRT->mutex_lock);
	if (BGP->views[view] == NULL) {
	    /* a new view */
#ifdef HAVE_IPV6
	    if (prefix->family == AF_INET6) {
		BGP->views[view] = New_View (128);
    	        sprintf (str, "BGP4+ view %d", BGP->views[view]->viewno);
	    }
	    else
#endif /* HAVE_IPV6 */
	    {
		BGP->views[view] = New_View (32);
    	        sprintf (str, "BGP view %d", BGP->views[view]->viewno);
	    }

	    BGP->views[view]->viewno = view;
	    BGP->views[view]->trace = trace_copy (BGP->trace);
    	    set_trace (BGP->views[view]->trace, TRACE_PREPEND_STRING, str, 0);

	}
        pthread_mutex_unlock (&MRT->mutex_lock);
	/* locking is not required for an assignment */
        pthread_mutex_lock (&peer->mutex_lock);
	peer->view = view;
        pthread_mutex_unlock (&peer->mutex_lock);
    }
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_view (uii_connection_t * uii, char *name,
			     int view)
{
    bgp_peer_t *peer;
    char str[MAXLINE];

    if (view >= MAX_BGP_VIEWS || view < 0) {
	config_notice (TR_ERROR, uii,
		       "Invalid view %d for BGP neighbor %s [0-%d]\n",
		       view, name, MAX_BGP_VIEWS - 1);
	Delete (name);
	return (-1);
    }

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    if (BGP->views[view]) {
	if (BGP->views[view]->family != peer->peer_addr->family) {
	    config_notice (TR_ERROR, uii, "view %d already defined for %s\n",
			   view, family2string (BGP->views[view]->family));
	    Delete (name);
	    return (-1);
	}

	trace (TR_PARSE, BGP->trace, "neighbor %s view %d\n", name, view);

        pthread_mutex_lock (&MRT->mutex_lock);
	if (BGP->views[view] == NULL) {
	    /* a new view */
#ifdef HAVE_IPV6
	    if (peer->peer_addr->family == AF_INET6) {
		BGP->views[view] = New_View (128);
    	        sprintf (str, "BGP4+ view %d", BGP->views[view]->viewno);
	    }
	    else
#endif /* HAVE_IPV6 */
	    {
		BGP->views[view] = New_View (32);
    	        sprintf (str, "BGP view %d", BGP->views[view]->viewno);
	    }

	    BGP->views[view]->viewno = view;
	    BGP->views[view]->trace = trace_copy (BGP->trace);
    	    set_trace (BGP->views[view]->trace, TRACE_PREPEND_STRING, str, 0);

	}
        pthread_mutex_unlock (&MRT->mutex_lock);
	/* locking is not required for an assignment */
        pthread_mutex_lock (&peer->mutex_lock);
	peer->view = view;
        pthread_mutex_unlock (&peer->mutex_lock);
    }
    Delete (name);
    return (1);
}


/* config_router_neighbor_weight
 */
static int
config_router_neighbor_weight (uii_connection_t * uii,
			       prefix_t * prefix, int weight)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    /* weight = BGP_PREF; */
    peer->default_pref = weight;

    trace (TR_PARSE, BGP->trace, "neighbor %s weight %d\n",
	   prefix_toa (prefix), weight);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_weight (uii_connection_t * uii,
			       char * name, int weight)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    /* weight = BGP_PREF; */
    peer->default_pref = weight;

    trace (TR_PARSE, BGP->trace, "neighbor %s weight %d\n", name, weight);
    Delete (name);
    return (1);
}


static int
config_router_neighbor_maximum_prefix (uii_connection_t * uii,
			       prefix_t * prefix, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    peer->maximum_prefix = num;

    trace (TR_PARSE, BGP->trace, "neighbor %s maximum-prefix %d\n",
	   prefix_toa (prefix), num);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_maximum_prefix (uii_connection_t * uii,
			                 char * name, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    peer->maximum_prefix = num;

    trace (TR_PARSE, BGP->trace, "neighbor %s maximum-prefix %d\n",
	   name, num);
    Delete (name);
    return (1);
}


/* config_router_neighbor_alias
 */
static int
config_router_neighbor_alias (uii_connection_t * uii,
			      prefix_t * prefix, prefix_t * prefix_alias)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s alias %s\n",
	   prefix_toa (prefix), prefix_toa (prefix_alias));

    if (peer->aliases == NULL)
	peer->aliases = LL_Create (0);

    LL_Add (peer->aliases, prefix_alias);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_router_neighbor_n_alias (uii_connection_t * uii, char *name,
			      prefix_t * prefix_alias)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s alias %s\n",
	   name, prefix_toa (prefix_alias));

    if (peer->aliases == NULL)
	peer->aliases = LL_Create (0);

    LL_Add (peer->aliases, prefix_alias);
    Delete (name);
    return (1);
}


static int
config_router_neighbor_option (uii_connection_t * uii, prefix_t *prefix,
			         char *option)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (option);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s\n",
	   prefix_toa (prefix), option);

    if (strcasecmp (option, "transparent-as") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_AS);
    else if (strcasecmp (option, "transparent-nexthop") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
    else if (strcasecmp (option, "passive") == 0) {
        BIT_SET (peer->options, BGP_CONNECT_PASSIVE);
    }
    else {
	assert (0);
    }
    Deref_Prefix (prefix);
    Delete (option);
    return (1);
}


static int
config_router_neighbor_n_option (uii_connection_t * uii, char *name,
			         char *option)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	Delete (option);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s\n", name, option);

    if (strcasecmp (option, "transparent-as") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_AS);
    else if (strcasecmp (option, "transparent-nexthop") == 0)
        BIT_SET (peer->options, BGP_TRANSPARENT_NEXTHOP);
    else if (strcasecmp (option, "passive") == 0) {
        BIT_SET (peer->options, BGP_CONNECT_PASSIVE);
    }
    else {
	assert (0);
    }
    Delete (name);
    Delete (option);
    return (1);
}


static int
config_router_neighbor_list_in_out (uii_connection_t * uii, prefix_t * prefix, 
				    char *op, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (op);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace,
	   "neighbor %s %s %d %s\n", prefix_toa (prefix), op, num, in_or_out);
    if (strcasecmp (op, "distribute-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
            peer->dlist_in = num;
	else
            peer->dlist_out = num;
    }
    else if (strcasecmp (op, "filter-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
    	    peer->flist_in = num;
	else
    	    peer->flist_out = num;
    }
    else {
	assert (0);
    }
    Deref_Prefix (prefix);
    Delete (op);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_n_list_in_out (uii_connection_t * uii, char *name, 
				      char *op, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	Delete (op);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace,
	   "neighbor %s %s %d %s\n", name, op, num, in_or_out);
    if (strcasecmp (op, "distribute-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
            peer->dlist_in = num;
	else
            peer->dlist_out = num;
    }
    else if (strcasecmp (op, "filter-list") == 0) {
	if (strcasecmp (in_or_out, "in") == 0)
    	    peer->flist_in = num;
	else
    	    peer->flist_out = num;
    }
    else {
	assert (0);
    }
    Delete (name);
    Delete (op);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_routemap_in_out (uii_connection_t * uii,
				  prefix_t * prefix, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s route-map %d %s\n",
	   prefix_toa (prefix), num, in_or_out);
    if (strcasecmp (in_or_out, "in") == 0)
        peer->route_map_in = num;
    else
        peer->route_map_out = num;
    Deref_Prefix (prefix);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_n_routemap_in_out (uii_connection_t * uii,
				       char * name, int num, char *in_or_out)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
        Delete (in_or_out);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s route-map %d out\n",
	   name, num);
    if (strcasecmp (in_or_out, "in") == 0)
        peer->route_map_in = num;
    else
        peer->route_map_out = num;
    Delete (name);
    Delete (in_or_out);
    return (1);
}


static int
config_router_neighbor_bgp4plus (uii_connection_t * uii,
				 prefix_t * prefix, char *version)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
        Delete (version);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s bgp4+ %s\n",
	   prefix_toa (prefix), version);

    BIT_SET (peer->options, BGP_BGP4PLUS);
    if (strcasecmp (version, "1") == 0 || strcasecmp (version, "new") == 0 ||
        strcasecmp (version, "rfc") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_01);
    else if (strcasecmp (version, "auto") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_AUTO);

    Deref_Prefix (prefix);
    Delete (version);
    return (1);
}


static int
config_router_neighbor_n_bgp4plus (uii_connection_t * uii,
				   char * name, char * version)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
        Delete (version);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s bgp4+ %s\n",
	   name, version);

    BIT_SET (peer->options, BGP_BGP4PLUS);
    if (strcasecmp (version, "1") == 0 || strcasecmp (version, "new") == 0 ||
        strcasecmp (version, "rfc") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_01);
    else if (strcasecmp (version, "auto") == 0)
	BIT_SET (peer->options, BGP_BGP4PLUS_AUTO);

    Delete (name);
    Delete (version);
    return (1);
}


static int
config_router_neighbor_time (uii_connection_t * uii,
			     prefix_t * prefix, char *timer, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	Delete (timer);
	return (-1);
    }
    trace (TR_PARSE, BGP->trace, "neighbor %s %s %d\n",
	   prefix_toa (prefix), timer, num);

    if (strcasecmp (timer, "holdtime") == 0) {
        peer->HoldTime_Interval = num;
	Timer_Set_Time (peer->timer_HoldTime, num);
    }
    else if (strcasecmp (timer, "keepalive") == 0) {
        peer->KeepAlive_Interval = num;
	Timer_Set_Time (peer->timer_KeepAlive, num);
    }
    else if (strcasecmp (timer, "connectretry") == 0) {
        peer->ConnectRetry_Interval = num;
	Timer_Set_Time (peer->timer_ConnectRetry, num);
    }
    else if (strcasecmp (timer, "starttime") == 0) {
        peer->Start_Interval = num;
	Timer_Set_Time (peer->timer_Start, num);
    }
    else {
	assert (0);
    }

    Delete (prefix);
    Delete (timer);
    return (1);
}


static int
config_router_neighbor_n_time (uii_connection_t * uii,
			       char * name, char *timer, int num)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    trace (TR_PARSE, BGP->trace, "neighbor %s %s %d\n",
	   name, timer, num);

    if (strcasecmp (timer, "holdtime") == 0) {
        peer->HoldTime_Interval = num;
	Timer_Set_Time (peer->timer_HoldTime, num);
    }
    else if (strcasecmp (timer, "keepalive") == 0) {
        peer->KeepAlive_Interval = num;
	Timer_Set_Time (peer->timer_KeepAlive, num);
    }
    else if (strcasecmp (timer, "connectretry") == 0) {
        peer->ConnectRetry_Interval = num;
	Timer_Set_Time (peer->timer_ConnectRetry, num);
    }
    else if (strcasecmp (timer, "starttime") == 0) {
        peer->Start_Interval = num;
	Timer_Set_Time (peer->timer_Start, num);
    }
    else {
	assert (0);
    }

    Delete (name);
    Delete (timer);
    return (1);
}


static int
trace_bgp_neighbor (uii_connection_t * uii, prefix_t *prefix)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    set_trace (peer->trace, TRACE_ADD_FLAGS, TR_ALL, NULL);
    Deref_Prefix (prefix);
    return (1);
}


static int
trace_bgp_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    set_trace (peer->trace, TRACE_ADD_FLAGS, TR_ALL, NULL);
    Delete (name);
    return (1);
}


static int
no_trace_bgp_neighbor (uii_connection_t * uii, prefix_t *prefix)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
	Deref_Prefix (prefix);
	return (-1);
    }

    set_trace (peer->trace, TRACE_DEL_FLAGS, TR_ALL, NULL);
    Deref_Prefix (prefix);
    return (1);
}


static int
no_trace_bgp_neighbor_n (uii_connection_t * uii, char *name)
{
    bgp_peer_t *peer;

    if ((peer = Find_BGP_Peer_ByID (name)) == NULL) {
	config_notice (TR_ERROR, uii, "No peer %s\n", name);
	Delete (name);
	return (-1);
    }

    set_trace (peer->trace, TRACE_DEL_FLAGS, TR_ALL, NULL);
    Delete (name);
    return (1);
}


static int
config_router_aggregate (uii_connection_t * uii, prefix_t * prefix, 
			 char *rest)
{
    u_long option;
    char opt[4][MAXLINE];
    int view = 0, n;

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	view = 1;
#endif /* HAVE_IPV6 */

    opt[0][0] = opt[1][0] = opt[2][0] = opt[3][0] = '\0';
    option = 0;
    if (rest && (n = parse_line (rest, "%s %s %s %s",
				 opt[0], opt[1], opt[2], opt[3])) >= 1) {
	int i;

	for (i = 0; i < n; i++) {

	    if (strcasecmp (opt[i], "as-set") == 0) {
	        BIT_SET (option, BGP_AGGOPT_AS_SET);
	    }
	    else if (strcasecmp (opt[i], "summary-only") == 0) {
	        BIT_SET (option, BGP_AGGOPT_SUMMARY_ONLY);
	    }
	    else if (i < n - 1 && strcasecmp (opt[i], "view") == 0
	        && isdigit (opt[i + 1][0])) {
	        view = atoi (opt[i + 1]);
		i++;
	    }
	}
    }

    if (rest)
	Delete (rest);

    if (BGP->views[view] == NULL) {
	config_notice (TR_TRACE, uii,
		    "CONFIG aggregate-address view %d not defined\n", view);
	Deref_Prefix (prefix);
	return (-1);
    }
    if (BGP->views[view]->family != prefix->family) {
	config_notice (TR_TRACE, uii,
		"CONFIG aggregate-address view %d family %s not %s\n", view,
		       family2string (BGP->views[view]->family),
		       family2string (prefix->family));
	Deref_Prefix (prefix);
	return (-1);
    }
    view_add_aggregate (BGP->views[view], prefix, option);
    config_notice (TR_TRACE, uii,
		   "CONFIG aggregate-address %s%s%s\n",
		   prefix_toax (prefix),
		   BIT_TEST (option, BGP_AGGOPT_AS_SET) ? " as-set" : "",
	 BIT_TEST (option, BGP_AGGOPT_SUMMARY_ONLY) ? " summary-only" : "");
    Deref_Prefix (prefix);
    return (1);
}


static int
no_config_router_aggregate (uii_connection_t * uii, prefix_t * prefix, 
			    char *rest)
{
    int view = 0;

#ifdef HAVE_IPV6
    if (prefix->family == AF_INET6)
	view = 1;
#endif
    if (rest) {
	parse_line (rest, "view %d", &view);
	Delete (rest);
    }

    if (BGP->views[view] == NULL) {
	config_notice (TR_ERROR, uii,
		 "no aggregate-address view %d not defined\n", view);
	Deref_Prefix (prefix);
	return (-1);
    }
    if (BGP->views[view]->family != prefix->family) {
	config_notice (TR_ERROR, uii,
		   "no aggregate-address view %d family %s not %s\n",
		       view,
		       family2string (BGP->views[view]->family),
		       family2string (prefix->family));
	Deref_Prefix (prefix);
	return (-1);
    }
    view_del_aggregate (BGP->views[view], prefix);
    config_notice (TR_TRACE, uii,
		   "CONFIG no aggregate-address %s view %d\n",
		   prefix_toa (prefix), view);
    Deref_Prefix (prefix);
    return (1);
}


static int
config_dump_bgp_common (uii_connection_t * uii, char *filename,
			char *unit, time_t * time)
{
    if (unit[0] && isdigit (unit[0])) {
	int i;
	u_int u;
	u = atoi (unit);
	for (i = 0; i < sizeof (unit) && unit[i] && isdigit (unit[i]); i++);
	if (i < sizeof (unit) && isalpha (unit[i])) {
	    if (strncasecmp (unit + i, "m", 1) == 0)
		u *= 60;
	    else if (strncasecmp (unit + i, "h", 1) == 0)
		u *= (60 * 60);
	}
	*time = u;
    }
    return (1);
}


static char *
time_to_unit (time_t t)
{
    char *buff;

    THREAD_SPECIFIC_STRAGE (buff);
    if ((t % 3600) == 0)
	sprintf (buff, "%ldh", t / 3600);
    else if ((t % 60) == 0)
	sprintf (buff, "%ldm", t / 60);
    else
	sprintf (buff, "%ld", t);
    return (buff);

}

static void
get_config_dump_bgp_updates ()
{
    config_add_output ("dump bgp updates %s %s\n", BGP->dump_update_form,
		       time_to_unit (BGP->dump_update_interval));
}


static void
get_config_dump_bgp_view (int id)
{
    config_add_output ("dump bgp view %d %s %s\n", id,
			   BGP->dump_route_form[id],
			   time_to_unit (BGP->dump_route_interval[id]));
}


static int
config_dump_bgp_updates (uii_connection_t * uii, char *filename, char *unit)
{
    time_t t = 0;

    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
	set_BGP (BGP_DUMP_UPDATE_FORM, filename, t, 0);
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp updates %s %d\n", filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_updates, NULL);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static int
config_dump_bgp_view (uii_connection_t * uii,
		      int id, char *filename, char *unit)
{
    time_t t = 0;

    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
        set_BGP (BGP_DUMP_ROUTE_FORM, id, filename, t, DUMP_ASCII, 0);
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp view %d %s %d\n", id, filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_view, id);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static void
get_config_dump_bgp_routes ()
{
  if (BGP->dump_route_type == DUMP_ASCII) 
    config_add_output ("dump bgp routes %s %s\n", BGP->dump_route_form[0],
		       time_to_unit (BGP->dump_route_interval[0]));
  else
    config_add_output ("dump-binary bgp routes %s %s\n", 
		       BGP->dump_route_form[0],
		       time_to_unit (BGP->dump_route_interval[0]));
}


static int
_config_dump_bgp_routes (uii_connection_t * uii, char *filename, 
			 char *unit, int dump_type)
{
    time_t t = 0;
    if (config_dump_bgp_common (uii, filename, unit, &t) >= 0 && filename[0]) {
	set_BGP (BGP_DUMP_ROUTE_FORM, 0, filename, t, dump_type, 0);
#ifdef HAVE_IPV6
	set_BGP (BGP_DUMP_ROUTE_FORM, 1, filename, t, dump_type, 0);
#endif /* HAVE_IPV6 */
	config_notice (TR_TRACE, uii,
		       "CONFIG dump bgp routes %s %d\n", filename, t);
	config_add_module (0, "bgp dump", get_config_dump_bgp_routes, NULL);
	Delete (filename);
	Delete (unit);
	return (1);
    }
    Delete (filename);
    Delete (unit);
    return (-1);
}


static int
config_dump_bgp_routes (uii_connection_t * uii, char *filename, char *unit) {
  
  return (_config_dump_bgp_routes (uii, filename, unit, DUMP_ASCII));
}


static int
config_dump_bgp_routes_binary (uii_connection_t * uii, char *filename, 
			       char *unit) {
  return (_config_dump_bgp_routes (uii, filename, unit, DUMP_BINARY));
}


static int
kill_peer (uii_connection_t * uii, char *whom)
{
    prefix_t *prefix;
    bgp_peer_t *peer;

    if (strcmp (whom, "*") == 0) {
	bgp_kill_all ();
	return (1);
    }
    else if (parse_line (whom, "%m", &prefix) >= 1) {

	if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
	    config_notice (NORM, uii, "No peer %s\n", prefix_toa (prefix));
	    return (-1);
	}

	return (bgp_kill_peer2 (peer));
    }
    return (-1);
}


void
config_bgp_init (void)
{
    set_uii (UII, UII_PROMPT, UII_CONFIG_ROUTER_BGP, "Router BGP> ", 0);

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp", show_bgp,
		      "Show BGP peers and their status");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show view %d", show_view,
		      "Show the BGP routing table for the view");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp routes",
		      show_bgp_routing_table, "Show BGP routing table");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show bgp regexp %S",
		      show_bgp_rt_regexp,
		      "Show BGP routes matching the regular expression");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "show bgp neighbors (%m|%n|*) errors",
		      show_bgp_neighbors_errors,
		      "Show BGP recent errors [with the peer]");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "show bgp neighbors (%m|%n|*) routes",
		      show_bgp_neighbors_routes,
		      "Show BGP routes sent [to the peer]");

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "clear bgp %sPeer_Name", kill_peer,
		      "Close BGP peering session");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "router bgp %d",
		      config_router_bgp,
		      "Enables BGP (BGP4+) routing protocol with AS number");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM,
		      "as-path access-list %d (permit|deny) %S",
		      config_ip_as_filter, "Defines AS path filter");

    /* move to uii_normal */
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp neighbor %m", trace_bgp_neighbor,
		      "Enable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp neighbor %m", 
		      no_trace_bgp_neighbor, "Disable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp neighbor %n", trace_bgp_neighbor_n,
		      "Enable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp neighbor %n", 
		      no_trace_bgp_neighbor_n, "Disable trace bgp neighbor");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "trace bgp view (*|inet|inet6|%d)", 
		      trace_bgp_view, "Enable trace bgp view");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, 
		      "no trace bgp view (*|inet|inet6|%d)", 
		      no_trace_bgp_view, "Disable trace bgp view");

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "trace bgp", trace_bgp,
		      "Enable trace BGP");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "no trace bgp", no_trace_bgp,
		      "Disable trace BGP");

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, 
		"dump bgp view %dview_number %sFilename %sInterval(e.g._1m)",
		 config_dump_bgp_view, "Dumps BGP routing table for the view");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "dump bgp routes %s %s",
		      config_dump_bgp_routes, "Dumps BGP routing table");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, 
		      "dump-binary bgp routes %sFilename %sInterval",
		      config_dump_bgp_routes_binary, 
		      "Dumps Binary BGP routing table");
    uii_add_command2 (UII_CONFIG, COMMAND_NORM, 
		      "dump bgp updates %sFilename %sInterval(e.g._1m)",
		      config_dump_bgp_updates, 
		      "Dumps BGP updates and state changes");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM, "network %m",
		      config_router_network_prefix, 
		      "Originates routes in BGP");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"redistribute (static|ospf|rip|ripng|direct|connected|kernel)",
		      config_redistribute,
		      "Redistribute route from the protocol");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"no redistribute (static|ospf|rip|ripng|direct|connected|kernel)",
		      no_config_redistribute,
		      "Not redistribute route from the protocol");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m remote-as %d",
		   config_router_neighbor_remoteas, "Adds an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no neighbor %m remote-as %d",
	     no_config_router_neighbor_remoteas, "Deletes an BGP neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m view %d", config_router_neighbor_view,
		      "Assigns a view for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m weight %d", config_router_neighbor_weight,
		      "Sets a weight for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m maximum-prefix %d", 
		       config_router_neighbor_maximum_prefix,
		      "Sets max number of prefixes for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m alias %n",
		      config_router_neighbor_alias, "Obsolete");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		  "neighbor %m (transparent-as|transparent-nexthop|passive)",
		      config_router_neighbor_option,
		      "Sets an option for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"neighbor %m (distribute-list|filter-list) %d (in|out)",
		      config_router_neighbor_list_in_out,
		      "Applies the filter for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m route-map %d (in|out)",
		      config_router_neighbor_routemap_in_out,
		      "Applies the route map for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m bgp4+ (0|1|old|new|rfc|auto)",
		      config_router_neighbor_bgp4plus,
		      "Specifies BGP4+ packet format");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		"neighbor %m (holdtime|keepalive|connectretry|starttime) %d",
		      config_router_neighbor_time,
		      "Sets timer for neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n peer %m %S",
		   config_router_neighbor_n_peer, "Adds an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no neighbor %n",
	     no_config_router_neighbor_n, "Deletes an BGP neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n remote-as %d",
		       config_router_neighbor_n_remoteas, 
		      "Sets AS number to the neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n view %d", config_router_neighbor_n_view,
		      "Assigns view for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n weight %d", config_router_neighbor_n_weight,
		      "Sets weight for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n maximum-refix %d", 
		       config_router_neighbor_n_maximum_prefix,
		      "Sets max number of prefixes for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n alias %n",
		      config_router_neighbor_n_alias, "Obsolete");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		  "neighbor %n (transparent-as|transparent-nexthop|passive)",
		      config_router_neighbor_n_option,
		      "Sets an option for the neighbor");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
	"neighbor %n (distribute-list|filter-list) %d (in|out)",
		      config_router_neighbor_n_list_in_out,
		      "Applies the filter for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n route-map %d (in|out)",
		      config_router_neighbor_n_routemap_in_out,
		      "Applies the route map for incoming or outgoing");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %n bgp4+ (0|1|old|new|rfc|auto)",
		      config_router_neighbor_n_bgp4plus,
		      "Specifies BGP4+ packet format");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		"neighbor %n (holdtime|keepalive|connectretry|starttime) %d",
		      config_router_neighbor_n_time,
		      "Sets timer for neighbor");

    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "aggregate-address %m %S", config_router_aggregate,
		      "Creates an aggregate entry to prefix");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "no aggregate-address %m %S", 
		       no_config_router_aggregate,
		      "Deletes an aggregate entry to prefix");
}
